"""
Users REST API module
"""

import logging
from flask import request
from flask_jwt_extended import jwt_required
from flask_restplus import Namespace, Resource

from portal.database.models import User
from portal.rest import rest_helpers
from portal.rest.schemas import courses_schema, groups_schema, password_change_schema, \
    reviews_schema, roles_schema, submissions_schema, user_schema, user_schema_reduced, \
    users_schema, projects_schema
from portal.service import general, permissions, users, auth
from portal.service.errors import ForbiddenError, PortalAPIError
from portal.service.permissions import check_component, check_sysadmin
from portal.service.users import create_user, delete_user, find_groups_filtered, find_reviews, \
    find_roles_filtered, find_users_filtered, update_user, \
    update_user_password

users_namespace = Namespace('users')

log = logging.getLogger(__name__)


@users_namespace.route('')
class UserList(Resource):
    @jwt_required
    # @users_namespace.response(200, 'List of users', model=users_schema)
    def get(self):
        client = auth.find_client()
        permissions.require_sysadmin(client)
        course_id = request.args.get('course')
        group_id = request.args.get('group')
        if group_id and not course_id:
            raise PortalAPIError(400,
                                 "Invalid filter combination at /get/users: "
                                 "missing course id.")

        filtered_users = find_users_filtered(
            course_id=course_id, group_id=group_id)
        return users_schema.dump(filtered_users)[0], 200

    @jwt_required
    # @users_namespace.response(201, 'Created user', model=user_schema)
    def post(self):
        client = auth.find_client()
        permissions.require_sysadmin(client)
        data = rest_helpers.parse_request_data(
            user_schema, action='create', resource='user')
        new_user = create_user(**data)
        return user_schema.dump(new_user)[0], 201


@users_namespace.route('/<string:uid>')
@users_namespace.param('uid', "User's id")
@users_namespace.response(404, 'User not found')
class UserResource(Resource):
    @jwt_required
    # @users_namespace.response(200, "User reduced info", model=user_schema_reduced)
    # @users_namespace.response(200, "User info", model=users_schema)
    def get(self, uid: str):
        client = auth.find_client()
        user = general.find_user(uid)
        # authorization
        schema = get_user_schema_based_on_permissions(client, user)
        return schema.dump(user)[0]

    @jwt_required
    @users_namespace.response(204, 'User updated')
    def put(self, uid: str):
        client = auth.find_client()
        user = general.find_user(uid)
        _require_user_access(client, user)

        data = rest_helpers.parse_request_data(
            user_schema, action='update', resource='user', partial=True
            )
        update_user(user, data, permissions.check_sysadmin(client))
        return '', 204

    @jwt_required
    @users_namespace.response(204, 'User deleted')
    def delete(self, uid: str):
        client = auth.find_client()
        permissions.require_sysadmin(client)
        user = general.find_user(uid)
        delete_user(user)
        return '', 204


@users_namespace.route('/<string:uid>/password')
@users_namespace.param('uid', "User's id")
@users_namespace.response(404, 'User not found')
class UserPassword(Resource):
    @jwt_required
    @users_namespace.response(204, 'User password updated')
    def put(self, uid: str):
        client = auth.find_client()
        user = general.find_user(uid)
        _require_user_access(client, user)
        data = rest_helpers.parse_request_data(
            password_change_schema, action='update', resource='user_password'
            )
        update_user_password(client, data, user)
        return '', 204


@users_namespace.route('/<string:uid>/submissions')
@users_namespace.param('uid', "User's id")
@users_namespace.response(404, 'User not found')
class UserSubmissionList(Resource):
    @jwt_required
    # @users_namespace.response(200, 'Submissions list', model=submissions_schema)
    def get(self, uid: str):
        client = auth.find_client()
        user = general.find_user(uid)

        course_id = request.args.get('course')
        project_ids = request.args.getlist('project')
        if project_ids and not course_id:
            raise PortalAPIError(400,
                                 f"Invalid filter combination at "
                                 f"/get/users/{uid}/submissions: missing course id.")

        submissions = get_submissions_based_on_permissions(
            client, user, course_id, project_ids)
        return submissions_schema.dump(submissions)[0]


def find_groups_intersection(user1: User, user2: User):
    other_groups = user2.groups
    res = [group for group in user1.groups if group in other_groups]
    return res


@users_namespace.route('/<string:uid>/roles')
@users_namespace.param('uid', "User's id")
@users_namespace.response(404, 'User not found')
class UserRoleList(Resource):
    @jwt_required
    # @users_namespace.response(200, 'User roles list', model=roles_schema)
    def get(self, uid: str):
        client = auth.find_client()
        user = general.find_user(uid)
        _require_user_access(client, user)
        course_id = request.args.get('course')

        roles = find_roles_filtered(user=user, course_id=course_id)
        return roles_schema.dump(roles)[0]


@users_namespace.route('/<string:uid>/courses')
@users_namespace.param('uid', "User's id")
@users_namespace.response(404, 'User not found')
class UserCourseList(Resource):
    @jwt_required
    # @users_namespace.response(200, 'Courses roles list', model=courses_schema)
    def get(self, uid: str):
        client = auth.find_client()
        user = general.find_user(uid)
        _require_user_access(client, user)
        return courses_schema.dump(user.courses)[0]


@users_namespace.route('/<string:uid>/projects')
@users_namespace.param('uid', "User's id")
@users_namespace.response(404, 'User not found')
class UserProjectsList(Resource):
    @jwt_required
    # @users_namespace.response(200, 'Courses roles list', model=courses_schema)
    def get(self, uid: str):
        client = auth.find_client()
        user = general.find_user(uid)
        # authorization
        _require_user_access(client, user)
        return projects_schema.dump(users.find_all_projects_for_user(user))[0]


@users_namespace.route('/<string:uid>/groups')
@users_namespace.param('uid', "User's id")
@users_namespace.response(404, 'User not found')
class UserGroupList(Resource):
    @jwt_required
    # @users_namespace.response(200, 'User groups list', model=groups_schema)
    def get(self, uid: str):
        client = auth.find_client()
        user = general.find_user(uid)
        # authorization
        _require_user_access(client, user)
        course_id = request.args.get('course')
        groups = find_groups_filtered(user=user, course_id=course_id)
        return groups_schema.dump(groups)[0]


@users_namespace.route('/<string:uid>/reviews')
@users_namespace.param('uid', "User's id")
@users_namespace.response(404, 'User not found')
class UserReviewList(Resource):
    @jwt_required
    # @users_namespace.response(200, 'User reviews list', model=reviews_schema)
    def get(self, uid: str):
        client = auth.find_client()
        user = general.find_user(uid)
        # authorization
        _require_user_access(client, user)

        reviews = find_reviews(user)
        return reviews_schema.dump(reviews)[0]


@users_namespace.route('/<string:uid>/permissions')
@users_namespace.param('uid', "User's id")
@users_namespace.response(404, 'User not found')
class UserEffectivePermissions(Resource):
    @jwt_required
    @users_namespace.response(200, 'Effective permissions')
    def get(self, uid: str):
        client = auth.find_client()
        user = general.find_user(uid)
        # authorization
        _require_user_access(client, user)

        course_id = request.args.get('course')
        perm = permissions.get_effective_permissions(user, course_id)
        return perm


#
# Helper functions
#


def get_user_schema_based_on_permissions(client, user):
    if check_component(component=client) or check_sysadmin(
            client) or client == user:
        return user_schema
    return user_schema_reduced


def get_submissions_based_on_permissions(client, user, course_id, project_ids):
    if course_id:
        return get_submissions_based_on_permissions_for_course(
            client, course_id, project_ids, user)
    if client.is_admin:
        return user.submissions
    if client == user:
        return user.submissions
    else:
        raise ForbiddenError(uid=client.id)


def get_submissions_based_on_permissions_for_course(
        client, course_id, project_ids, user):
    course = general.find_course(course_id)
    checks = [
        permissions.check_component(component=client),
        permissions.check_sysadmin(client),
        permissions.check_user(client, course, ['read_submissions_all']),
        permissions.check_user(client, course, ['read_submissions_groups']) and
        find_groups_intersection(client, user),
        permissions.check_user(
            client, course, ['read_submissions_own']) and client == user
        ]
    permissions.require_any_check(client, checks)

    return users.find_submissions_filtered(
        user, course_id=course_id, project_ids=project_ids)


def _require_user_access(client, user):
    checks = [
        permissions.check_component(component=client),
        permissions.check_sysadmin(client),
        client == user
        ]
    permissions.require_any_check(client, checks=checks)
