import json
import pytest

from portal.database.models import Course, Role, User
from portal.service.errors import PortalAPIError
from . import utils

"""Basic tests to verify functionality of the API for User.

These tests do not mock the database or care about authentication/authorization.

"""


def test_create(client):
    users_before = len(User.query.all())
    request_dict = {
        "uco": 445574,
        "email": "baz@foo.com",
        "username": "xuser",
    }
    request_json = json.dumps(request_dict)
    response = utils.make_request(client, '/users', data=request_json,
                                  headers={"content-type": "application/json"},
                                  method='post')
    assert response.status_code == 201
    assert response.mimetype == 'application/json'

    resp_new_user = utils.extract_data(response)
    db_users = User.query.all()
    assert len(db_users) == users_before + 1
    utils.assert_user_in(db_users, resp_new_user)


def test_create_extra_key(client):
    users_before = len(User.query.all())
    request_dict = {
        "uco": 456789,
        "email": "baz@foo.com",
        "username": "xuser",
        "redundant key": True,
    }

    request_json = json.dumps(request_dict)
    response = utils.make_request(client, '/users', data=request_json,
                                  headers={"content-type": "application/json"},
                                  method='post')
    assert response.status_code == 201
    assert response.mimetype == 'application/json'

    resp_new_user = utils.extract_data(response)
    db_users = User.query.all()
    assert len(db_users) == users_before + 1
    utils.assert_user_in(db_users, resp_new_user)


def test_post_data_error_invalid_value(client):
    users_before = len(User.query.all())
    request_dict = {
        "uco": "should be a number",
        "email": "baz@foo.com",
        "username": "xuser",
    }
    request_json = json.dumps(request_dict)
    response = utils.make_request(client, '/users', data=request_json,
                                  headers={"content-type": "application/json"},
                                  method='post')
    assert response.status_code == 400
    assert response.mimetype == 'application/json'

    db_users = User.query.all()
    assert len(db_users) == users_before


def test_post_data_error_missing_uco_for_user(client):
    users_before = len(User.query.all())
    request_dict = {
        "email": "baz@foo.com",
        "username": "xuser",
    }

    request_json = json.dumps(request_dict)
    response = utils.make_request(client, '/users', data=request_json,
                                  headers={"content-type": "application/json"},
                                  method='post')
    assert response.status_code == 400
    assert response.mimetype == 'application/json'

    db_users = User.query.all()
    assert len(db_users) == users_before


def test_list(client):
    users_before = len(User.query.all())
    response = utils.make_request(client, '/users', method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    users = utils.extract_data(response)

    assert len(users) == users_before
    db_users = User.query.all()
    utils.assert_user_in(db_users, users[0])


def test_list_filter_course(client):
    cpp = Course.query.filter_by(codename="testcourse1").first()
    users_before = [user for user in User.query.all() if cpp in user.courses]

    response = utils.make_request(
        client, '/users?course=testcourse1', method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    users = utils.extract_data(response)
    assert len(users) == len(users_before)

    utils.assert_user_in(users_before, users[0])


def test_list_filter_group(client):
    cpp = Course.query.filter_by(codename="testcourse1").first()
    g = [group for group in cpp.groups if group.name == "seminar01"][0]
    users_before = g.users

    response = utils.make_request(
        client, '/users?course=testcourse1&group=seminar01', method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    users = utils.extract_data(response)
    assert len(users) == len(users_before)

    utils.assert_user_in(users_before, users[0])


def test_list_filter_missing_course(client):
    response = utils.make_request(client, '/users?group=g1', method='get')
    assert response.status_code == 400


def test_read(client):
    user = User.query.all()[0]
    response = utils.make_request(client, f'/users/{user.id}', method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    data = utils.extract_data(response)
    utils.assert_user(user, data)


def test_update(client):
    users_before = len(User.query.all())
    user = User.query.all()[0]
    request_dict = {
        "name": "new_name",
        "email": "new_email@domain.net",
        "uco": 456789,
        "is_admin": True,
    }
    request_json = json.dumps(request_dict)
    response = utils.make_request(client, f'/users/{user.id}', data=request_json,
                                  headers={"content-type": "application/json"}, method='put')

    assert response.status_code == 204
    assert response.mimetype == 'application/json'
    updated_entity = User.query.filter_by(id=user.id).first()
    assert updated_entity.name == request_dict["name"]
    assert updated_entity.email == request_dict["email"]
    assert updated_entity.uco == request_dict["uco"]
    assert updated_entity.is_admin == request_dict["is_admin"]
    assert len(User.query.all()) == users_before


def test_delete(client):
    users_before = len(User.query.all())
    user = User.query.all()[0]

    response = utils.make_request(client, f"/users/{user.id}", method='delete')
    assert response.status_code == 204
    assert response.mimetype == 'application/json'

    db_users = User.query.all()
    assert len(db_users) == users_before - 1
    assert user not in db_users


def test_read_submissions(client):
    user = User.query.filter_by(username="student1").first()
    user_submissions = user.submissions

    response = utils.make_request(
        client, f"/users/{user.id}/submissions", method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'
    # check if the result is paginated?
    submissions = utils.extract_data(response)
    assert len(submissions) == len(user_submissions)
    utils.assert_submission_in(user_submissions, submissions[0])


def test_read_submissions_filter_course(client):
    user = User.query.filter_by(username="student2").first()
    user_submissions = [
        s for s in user.submissions if s.project.course.codename == "testcourse1"]

    response = utils.make_request(client, f"/users/{user.id}/submissions?course=testcourse1",
                                  method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    submissions = utils.extract_data(response)
    assert len(submissions) == len(user_submissions)
    utils.assert_submission_in(user_submissions, submissions[0])


def test_read_submissions_filter_course_project(client):
    user = User.query.filter_by(username="student1").first()
    user_submissions = [s for s in user.submissions if
                        s.project.name == "HW01" and s.project.course.codename == "testcourse1"]

    response = utils.make_request(client,
                                  f"/users/{user.id}/submissions?course=testcourse1&project=HW01",
                                  method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    submissions = utils.extract_data(response)
    assert len(submissions) == len(user_submissions)
    utils.assert_submission_in(user_submissions, submissions[0])


def test_read_submissions_filter_missing_course(client):
    user = User.query.filter_by(username="student1").first()
    path = f"/users/{user.id}/submissions?project=p1"
    response = utils.make_request(client, path, method='get')
    assert response.status_code == 400


def test_read_roles(client):
    user = User.query.filter_by(username="student2").first()
    roles = user.roles

    response = utils.make_request(
        client, f"/users/{user.id}/roles", method='get')

    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    user_roles = utils.extract_data(response)
    assert len(user_roles) == len(roles)
    utils.assert_role_in(roles, user_roles[0])


def test_read_roles_filter_course(client):
    user = User.query.filter_by(username="student2").first()
    roles = [role for role in user.roles if role.course.codename == "testcourse1"]
    response = utils.make_request(client, f"/users/{user.id}/roles?course=testcourse1",
                                  method='get')

    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    data = str(response.get_data().decode("utf-8"))
    user_roles = json.loads(data)
    assert len(user_roles) == len(roles)
    utils.assert_role_in(roles, user_roles[0])


def test_read_courses(client):
    user = User.query.filter_by(username="student2").first()
    courses = user.courses

    response = utils.make_request(
        client, f"/users/{user.id}/courses", method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    user_courses = utils.extract_data(response)
    assert len(user_courses) == len(courses)

    roles = Role.query.all()
    roles_filtered = [role for role in roles if user in role.users]

    courses2 = [role.course for role in roles_filtered]
    assert len(courses2) == len(courses)
    utils.assert_course_in(courses2, user_courses[0])


def test_read_groups(client):
    user = User.query.filter_by(username="student2").first()
    groups = user.groups
    response = utils.make_request(
        client, f"/users/{user.id}/groups", method='get')

    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    user_groups = utils.extract_data(response)
    assert len(user_groups) == len(groups)
    utils.assert_group_in(groups, user_groups[0])


def test_read_groups_filter_course(client):
    user = User.query.filter_by(username="student2").first()
    groups = [
        group for group in user.groups if group.course.codename == "testcourse1"]
    response = utils.make_request(client, f"/users/{user.id}/groups?course=testcourse1",
                                  method='get')

    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    user_groups = utils.extract_data(response)
    assert len(user_groups) == len(groups)

    utils.assert_group_in(groups, user_groups[0])


def test_user_reviews(client):
    user = User.query.filter_by(username="teacher1").first()
    reviews = []
    for item in user.review_items:
        if item.review not in reviews:
            reviews.append(item.review)

    response = utils.make_request(
        client, f"/users/{user.id}/reviews", method='get')

    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    user_reviews = utils.extract_data(response)
    assert len(user_reviews) == len(reviews)
    # the test is incomplete - only print
    print(f"user_reviews: {user_reviews}")
